#!/usr/bin/perl
#==============================================================================
#==============================================================================
#
# Name:		hosts.pl
#
#		Host Oriented Security Test Suite
#
# Description:	This script performs a kernel program compliance test as 
# 		specified test sequence creating a numeric evaluation of 
#		percent passed or failed.
#
# History:	REV      DATE      PROGRAMMER   DESCRIPTION
#               ===   =========   ============  ==============================
#               1.0   23 Jul 01   Finegan       Original
#		1.1   14 Aug 01   Finegan       Modify how comments display
#		1.2   03 Oct 01   Finegan       Add test bypass capability
#		1.3   30 Oct 01   Finegan       Add auto-create SKIPTEST flag
#						capability
#		1.4   04 Dec 01   Finegan       Remove excessive protection
#						logic that breaks under HP-UX
#		1.4.x 31 Jan 02	  Finegan       Change default execution order
#		1.5   04 Feb 02   Finegan       Add OS plugin tree support
#
# Corporation:  The MITRE Corporation, 1820 Dolley Madison, McLean, VA 22102
#
#==============================================================================
#==============================================================================
#
#
# Key Header Definitions
# =============================================================================
$version = "Version 1.5.0.0    Released 04 February, 2002";
$originator = "The MITRE Corporation";
$program = "Center for Integrated Intelligence Systems";
$project = "Network Security Engineering";
$utility = "Host Oriented Security Test Suite (HOSTS)";
$default_test_list = "version_series ia_series audit_series misc_series os_series dac_series"; 
#
#
# Define Headers
# =============================================================================
format TOP =

.
format HEADER_DATA =

 @|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$originator
 @|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$program
 @|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$project

 @|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$utility
 @|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
$version


 Generated on:                    Output Mode:                  Generated by:
 @<<<<<<<<<<<<<<<<<<<<<    @|||||||||||||||||||||||||           @>>>>>>>>>>>>
$generated_on,$verbose_mode,$generated_by

 Test Input File:                                              Test Run Time:
 @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< @>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
$test_name,$run_date

 ==============================================================================
 ==============================================================================

.
format SEPERATOR =

********************************************************************************
********************************************************************************
.
format RESULTS_DATA =
 
    Test: ^<<<<<<<<<<<<<<<<<<<<<<<   Test Description:
$test_number
    ===========================================================================
    CUSP:       ^<<<<<<<<<<<<<<<<<   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$cusp_reference, $description
    Req. Spec.: ^<<<<<<<<<<<<<<<<<   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$srs_reference, $description
              ~~^<<<<<<<<<<<<<<<<< ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$srs_reference, $description

      Test Command(s):  
      -------------------------------------------------------------------------
    ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$test_command
  
      <------------------------------- Results ------------------------------->
      Expected           Actual   Data Returned:
      -------------------------   ---------------------------------------------
    ~~^<<<<<<<<<< ~~^>>>>>>>>>> ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$expected_result, $actual_result, $result
  
      @|||||||||||||||||||||| Test Status: @<<<<< @|||||||||||||||||||||| 
$exclamation, $pass, $exclamation
 
.
format SECTION_SEPARATION =

 ******************************************************************************
 ******************************************************************************
 ***                                                                        ***
 ***~~^|||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||  ***
$section_header
 ***                                                                        ***
 ******************************************************************************
 ******************************************************************************

.
format SECTION_SEPARATION_BRIEF =
 Now Processing: ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$section_header
               ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$section_header
.
format COMMENT_BRIEF =
                 ----
                 ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$section_header
               ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$section_header
.
format TRAILER_DATA_1 =

 ==============================================================================
 ==============================================================================

 Start Date/Time:                                         Completion Date/Time:
 @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<              @>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
$run_date,$end_date

    <------------------------- Tests Run Summary -------------------------->
         Executed             Passed (@>>%)               Failed (@>>%)
$percent_passed, $percent_failed
    ========================================================================
    @|||||||||||||||||     @||||||||||||||||||         @||||||||||||||||||
$total_count, $pass_count, $fail_count

    Failed Tests:     ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$failed_tests
                    ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$failed_tests

.
format TRAILER_DATA_2 =
    Bypassed Tests:   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$bypass_list
                    ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$bypass_list

.
format TRAILER_DATA_3 =
    <----------------------- Requirements Summary ------------------------->
         Executed             Passed (@>>%)               Failed (@>>%)
$req_percent_passed, $req_percent_failed
    ========================================================================
    @|||||||||||||||||     @||||||||||||||||||         @||||||||||||||||||
$req_total_count, $req_pass_count, $req_fail_count


    Req. Status    Requirement List
    ============== =========================================================
    Met:           ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$validated_requirements
                 ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$validated_requirements
    Partially Met: ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$partial_requirements
                 ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$partial_requirements
    Not Met:       ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$failed_requirements
                 ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$failed_requirements


         Met:            Requirements where ALL tests associated with
                         the specified requirement PASSED.
         Partially Met:  Requirements where ONE OR MORE tests associated
                         with the specified requirement FAILED.
         Not Met:        Requirements where ALL tests associated with 
                         the specified requirement FAILED.

.
format SKIPTEST_MESSAGE =

                            N O T I C E ! ! !

      It is ill advised to automatically skip a test simply because 
      the test is failing.

      This routine is updating the SKIPTEST variable so that all failed
      tests will be skipped in future executions of this test series.  
      Before this action is taken, the following should be checked:

      1.  Do any of the tests that are now failing need to be modified?
          Consider (and fix as appropriate) the following common test 
          step failure causes:
          a.  Relocation of the tested file
          b.  Change in parameter name or value
          c.  Addlitional (equally secure) returned values are valid
          d.  Were the software and/or configuration files associated
              with a tested application recently updated?

      2.  If an application is no longer installed, tests pertaining
          to the application may no longer be valid in context of
          this series.  If this is true, the series should be updated
          accordingly.  (Be sure to retain the original series for
          regression testing.)

      3.  If this is a port from one operating system to another, or 
          from one release to another, was the port done correctly?

      4.  If the test step compares a list generated on the system
          under test against a baseline list, is the baseline list
          correct and/or up-to-date?  Do additional filters need to
          be integrated to eliminate more chaff before the compare is
          performed?

      It all boils down to this.  Before a test is simply skipped, make
      sure the test is correct first.  Failing to do this is lessens the
      the validity of the testing effort and could impact the validity
      of the test results.


As a precaution, this utility will make a backup copy of the test series:
Original File:   @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$original_file
Updated File:    @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$updated_file
Old Skip List:   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$old_bypass_list
               ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$old_bypass_list
New Skip List:   ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$new_bypass_list
               ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$new_bypass_list


The following table lists all of the tests from the original SKIPTEST 
definition that either now fail or now pass.  Each failing test should
be checked.  Rerun the automatic generation if any changes are made to
any of the tests.

    Test Step           P / F  Test Step Description
    ==================  =====  ===============================================
.
format SKIPTEST_DATA =
    @<<<<<<<<<<<<<<<<<  ^<<<<  ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$test_step, $test_state, $test_description
                             ~~^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$test_description
.
#
#
# Library Requirements
# =============================================================================
require 'ctime.pl';
use File::Basename;
#
#
# Save command line options
# =============================================================================
@command_options = join(' ', $0, @ARGV);
#
#
# Extract program name and directory
# =============================================================================
$program_name = basename $0;
$program_directory = dirname $0;
#
#
# Fetch debug mode flag
# =============================================================================
$debug_mode = $ENV{'HOSTS_DEBUG_MODE'};
#
#
# Fetch OS component to test path
# =============================================================================
$os_component = $^O;
if ( "$os_component" eq "" ) {
   $os_component = "default";
}
#
#
# Fetch path to tests
# =============================================================================
$test_path = $ENV{'HOSTS_TEST_PATH'};
if ( "$test_path" ne "" ) {
   $test_directory = $test_path;
}
else {
   $test_directory = "$program_directory/$os_component";
}
#
#
# Set up for command line argument processing
# =============================================================================
$verbose = "false";
$verbose_mode = "Failures Only";
$autogen_skiptest = "false";
$autogen_mode = "";
$test_list = "";
#
#
# Process command line arguments 
# =============================================================================
while ( $ARGV[0] ) {
# If verbose qualifier entered, tag verbose display
# =================================================
   if ( "$ARGV[0]" eq "-verbose") {
      $verbose = "true";
      $verbose_mode = "Verbose";
   }
# If automatic generation qualifier entered, enable automatic SKIPTEST build
# ===================================================
   elsif ( "$ARGV[0]" eq "-autogen") {
      $autogen_skiptest = "true";
      $autogen_mode = "/AutoGen";
   }
# Else build lists of tests to be executed
# ===================================================
   else {
      if ( "$test_list" eq "" ) { $test_list = "$ARGV[0]"; }
      else { $test_list = "$test_list $ARGV[0]"; }
   }
# Shift to process next argument
# ===================================================
   shift @ARGV;
}
# If null, set to default
# ===================================================
if ( "$test_list" eq "" ) {
   $test_list = "$default_test_list";
}
# Merge verbose and automatic generation mode messages
# ===================================================
$verbose_mode = "$verbose_mode$autogen_mode";
#
#
# Set up global definitions
# =============================================================================
@uname = split(/\s+/,`/bin/uname -a`);
$generated_on = $uname[1];
#
#
# Fetch name of user generating report
# =============================================================================
   $generated_by = $ENV{"USER"};
   if ( "$generated_by" eq "" ) {
      $generated_by = $ENV{"LOGNAME"};
   }
#
#
# Define required environmental variables
# =============================================================================
$ENV{'PATH'}="$test_directory/plugins:/usr/bin:/bin";
$ENV{'HOSTS_PLUGINS'}="$test_directory/plugins";
$ENV{'HOSTS_BASELINES'}="$test_directory/baseline";
#
#
# Reset global counters
# =============================================================================
$global_pass_count = 0;
$global_req_pass_count = 0;
$global_fail_count = 0;
$global_req_fail_count = 0;
$global_total_count = 0;
$global_req_total_count = 0;
$series = 0;
$global_passed_requirements = "";
$global_failed_requirements = "";
$global_failed_tests = "";
#
#
# Loop through specified tests
# =============================================================================
foreach $current_test ( split(' ',$test_list)) {
   $test_source = "$test_directory/$current_test";
#
#
# If not first test in loop, display seperator
# =============================================================================
   $series++;
   if ( $series > 1 ) {
      $~ = SEPERATOR;
      $= = 1000000;
      write;
   }
#
#
# Open test file skipping on error
# =============================================================================
   $! = 0;
   open (CURRENT_TEST,"<$test_source") or 
      warn "Error - Test file $test_source not found: $!\n";
#
#
# On failed open, skip further processing
# =============================================================================
   if ( "$!" ne "" ) {
      $series--;
      next;
   }
#
#
# Fetch run date and time, name of test
# =============================================================================
   $run_date=&ctime(time);
   $test_name=$test_source;
#
#
# Print header for test 
# =============================================================================
   $^ = TOP;
   $~ = HEADER_DATA;
   $= = 1000000;
   write;
#
#
# Loop through test items in list
# =============================================================================
   $got_data = "false";
   $pass_count = 0;
   $req_pass_count = 0;
   $fail_count = 0;
   $req_fail_count = 0;
   $total_count = 0;
   $req_total_count = 0;
   $bypass_list = "";
   $bypass_flag = "false";
   $redundant_skiptest_commands = "false";
   $failed_tests = "";
   $passed_requirements = "";
   $failed_requirements = "";
   while (<CURRENT_TEST>) {
#
#
# Clean up data - remove blank space, blank lines, comments
# =============================================================================
      chomp;
      s/^\s+//;
      s/\s+$//;
      next if (/^\#/ or /^\s*$/);
#
#
# Extract test components
# =============================================================================
      ($test_number, $description, $test_command, $expected_result, 
         $cusp_reference, $srs_reference, @garbage) = split(/\;/,$_);
      (@expected_table) = split(/\|/,$expected_result);
      $size_of_expected_table = @expected_table;
#
#
# Test for section header flag
# =============================================================================
      ($section_tag, $section_header, @garbage) = split(/\=/,$test_number); 
      if ( "$section_tag" eq "SECTION" ) {
         if ( "$verbose" eq "true" ) { $~ = SECTION_SEPARATION; }
         else { $~ = SECTION_SEPARATION_BRIEF; }
         write;
      }
#
#
# Test for comment flag
# =============================================================================
      elsif ( "$section_tag" eq "COMMENT") {
         $~ = COMMENT_BRIEF;
         write;
      }
#
#
# Test for comment flag
# =============================================================================
      elsif ( "$section_tag" eq "SKIPTEST") {
         if ( "$bypass_flag" eq "true" ) {
            print "\nERROR: Multiple SKIPTEST lines are not allowed!!\n";
            print "Redundant SKIPTEST line will be ignored.\n\n";
            $redundant_skiptest_commands = "true";
         }
         else {
            if ( "$autogen_skiptest" ne "true" ) {
               $bypass_list = join(', ', split (', ', $section_header));
            }
            else {
               $original_bypass_list = 
                  join(', ', split (', ', $section_header));
               if ( "$original_bypass_list" eq "" ) {
                  $original_bypass_list = "None";
               }
            }
            $bypass_flag = "true";
         }
      }
#
#
# Check for test bypass
# =============================================================================
      else {
         if ( $bypass_list !~ /$test_number/  ) {
#
#
# Perform test
# =============================================================================
            $command = "$test_command";
            $result = "";
            $size = run_command ( ) ;
            if ( $size > 0 ) {
               ($action_status, @recorded_data) = @command_results;
               $actual_result = $action_status;
               chomp( $actual_result );
               $result = join(' ', @recorded_data);
            }
            else {
               $actual_result = "Execution Error";
               $result = "Unable to perform test properly. Does the command specified exist?  Is it constructed properly?  Are all parameters and qualifiers properly defined?"
            }
            if ( $debug_mode ) {
               print "Returned result: @command_results \n";
            }
#
#
# Determine pass or fail
# =============================================================================
            $pass = "FAILED";
            if ( $size_of_expected_table > 1 ) {
               foreach $expectation (sort @expected_table) {
                  if ( "$actual_result" eq "$expectation"  && 
                  "$pass" ne "PASSED" ) {
                     $pass = "PASSED";
                     $expected_result = "$expectation";
                  }
               }
            }
            else {
               if ( "$expected_result" eq "$actual_result" ) {
                  $pass = "PASSED";
               }
            }
#
#
# Set counters accordingly
# =============================================================================
            $total_count++;
            $global_total_count++;
            if ( "$pass" eq "PASSED" ) { 
               $global_pass_count++; 
               $pass_count++; 
               $exclamation = "                     ";
            }
            else { 
               $global_fail_count++; 
               $fail_count++; 
               $exclamation = "!  !  !  !  !  !  !  ";
#
#
# Record failed test numbers
# =============================================================================
               if ( "$failed_tests" eq "" ) { $failed_tests = "$test_number"; }
               else { $failed_tests = "$failed_tests $test_number"; }
               if ( "$global_failed_tests" eq "" ) {
                  $global_failed_tests = "$test_number";
               }
               elsif ( $global_failed_tests !~ /$test_number/ &&
               global_failed_tests !~ /$test_number$/ ) {
                  $global_failed_tests = "$global_failed_tests $test_number";
               }
            }
#
#
# If applicable, record passed and failed SRS sections
# =============================================================================
            if ( "$srs_reference" ne "" ) {
               $global_req_total_count++;
               $req_total_count++;
               if ( "$pass" eq "PASSED" ) { 
                  $global_req_pass_count++; 
                  $req_pass_count++; 
                  foreach $requirement ( split(/, /,$srs_reference)) {
                     if ( "$passed_requirements" eq "" ) {
                        $passed_requirements = "$requirement";
                     }
                     elsif ( $passed_requirements !~ /$requirement/ &&
                     passed_requirements !~ /$requirement$/ ) {
                        $passed_requirements = 
                           "$passed_requirements $requirement";
                     }
                     if ( "$global_passed_requirements" eq "" ) {
                        $global_passed_requirements = "$requirement";
                     }
                     elsif ( $global_passed_requirements !~ /$requirement/ &&
                     $global_passed_requirements !~ /$requirement$/ ) {
                        $global_passed_requirements = 
                           "$global_passed_requirements $requirement";
                     }
                  }
               }
               else { 
                  $global_req_fail_count++; 
                  $req_fail_count++; 
                  foreach $requirement ( split(/, /,$srs_reference)) {
                     if ( "$failed_requirements" eq "" ) {
                        $failed_requirements = "$requirement";
                     }
                     elsif ( $failed_requirements !~ /$requirement/ &&
                     failed_requirements !~ /$requirement$/ ) {
                        $failed_requirements = 
                           "$failed_requirements $requirement";
                     }
                     if ( "$global_failed_requirements" eq "" ) {
                        $global_failed_requirements = "$requirement";
                     }
                     elsif ( $global_failed_requirements !~ /$requirement/ &&
                     global_failed_requirements !~ /$requirement$/ ) {
                        $global_failed_requirements = 
                           "$global_failed_requirements $requirement";
                     }
                  }
               }
            }
#
#
# Display results
# =============================================================================
            if ( "$verbose" eq "true" || "$pass" eq "FAILED" ) {
               $~ = RESULTS_DATA;
               write;
            }
         }
         else {
            if ( "$verbose" eq "true" ) {
               $exclamation = "*  *  *  *  *  *  *  ";
               $actual_result = "";
               $result = "";
               $pass = "BYPASS";
               $~ = RESULTS_DATA;
               write;
            }
         }
      }
   }
#
#
# Close test file
# =============================================================================
   close (CURRENT_TEST);
#
#
# Save failed test list for use by SKIPTEST autogen
# =============================================================================
   if ( "$autogen_skiptest" eq "true" ) { $failed_test_list = $failed_tests; }
#
#
# Display test series data
# =============================================================================
   &process_and_display_results;
#
#
# Process autobuild SKIPTEST variable
# =============================================================================
   &test_to_replace_skiptest;
#
#
# Return for next test in list
# =============================================================================
}
#
#
# If more than one test series, process global stats
# =============================================================================
if ( $series > 1 ) {
#
#
# Display separator and multi-test header
# =============================================================================
   $~ = SEPERATOR;
   $= = 1000000;
   write;
   print "\n                    Multi-Test Global Requirements Summary\n";
#
#
# Set up for subroutine call
# =============================================================================
   $pass_count = $global_pass_count;
   $fail_count = $global_fail_count;
   $total_count = $global_total_count;
   $req_total_count = $global_req_total_count;
   $req_pass_count = $global_req_pass_count;
   $req_fail_count = $global_req_fail_count;
   $passed_requirements = $global_passed_requirements;
   $failed_requirements = $global_failed_requirements;
   $failed_tests = $global_failed_tests;
#
#
# Display global data
# =============================================================================
   &process_and_display_results;
}
#
#
# Done
# =============================================================================
exit ;
#
#==============================================================================
#==============================================================================
#
#                 Subroutine to compute percentages
#
#==============================================================================
#==============================================================================
sub compute_percentages {
#
#
# Compute stats
# =============================================================================
   if ( $total_count > 0 ) {
      $percent_passed = int ( $pass_count * 100 / $total_count + .5 );
      $percent_failed = int ( $fail_count * 100 / $total_count + .5 );
      if ( $req_total_count > 0 ) {
         $req_percent_passed =
            int ( $req_pass_count * 100 / $req_total_count + .5 );
         $req_percent_failed =
            int ( $req_fail_count * 100 / $req_total_count + .5 );
      }
      else {
         $req_percent_passed = 0;
         $req_percent_failed = 0;
      }
   }
   else {
      $percent_passed = 0;
      $percent_failed = 0;
      $req_percent_passed = 0;
      $req_percent_failed = 0;
   }
}
#
#==============================================================================
#==============================================================================
#
#         Subroutine to process and display results summary
#
#==============================================================================
#==============================================================================
sub process_and_display_results {
#
#
# Compute percentages
# =============================================================================
   &compute_percentages;
#
#
# Fetch end time
# =============================================================================
   $end_date=&ctime(time);
#
#
# Sort failed summary and test data
# =============================================================================
   $failed_requirements = join(', ', sort( split(' ',$failed_requirements) ) );
   $failed_tests = join(', ', sort( split(' ',$failed_tests) ) );
#
#
# Compute passed requirement list based on which requirements failed
# =============================================================================
   $validated_requirements = "";
   $partial_requirements = "";
   foreach $requirement ( sort( split(/ /, $passed_requirements) ) ) {
      if ( $failed_requirements !~ /$requirement/ ) {
         if ( "$validated_requirements" eq "" ) {
            $validated_requirements = "$requirement";
         }
         else {
            $validated_requirements = "$validated_requirements, $requirement";
         }
      }
      else {
         if ( "$partial_requirements" eq "" ) {
            $partial_requirements = "$requirement";
         }
         else {
            $partial_requirements = "$partial_requirements, $requirement";
         }
      }
   }
#
#
# Remove any partially met requirements from the failed list
# =============================================================================
   $shortened_requirements = "";
   foreach $requirement ( sort( split(/, /, $failed_requirements) ) ) {
      if ( $partial_requirements !~ /$requirement/ ) {
         if ( "$shortened_requirements" eq "" ) {
            $shortened_requirements = "$requirement";
         }
         else {
            $shortened_requirements = "$shortened_requirements, $requirement";
         }
      }
   }
   $failed_requirements = $shortened_requirements;
#
#
# Adjust for null strings
# =============================================================================
   if ("$failed_tests" eq "") { $failed_tests = "None"; }
   if ("$failed_requirements" eq "") { $failed_requirements = "None"; }
   if ("$partial_requirements" eq "") {
      $partial_requirements = "Not Applicable"; }
   if ("$validated_requirements" eq "") { $validated_requirements = "None"; }
#
#
# Display summary
# =============================================================================
   $~ = TRAILER_DATA_1;
   write;
   if ( "$verbose" eq "true" ) {
      $~ = TRAILER_DATA_2;
      write;
   }
   $~ = TRAILER_DATA_3;
   write;
}
#
#==============================================================================
#==============================================================================
#
#              Command execution subroutine
#
#==============================================================================
#==============================================================================
sub run_command {
   local ( @ls_cmd, @which_cmd, $run_command, $test_string, $plugin, 
      @qualifiers, $return_str, $close_return_code );
   $close_return_code = 0;
   open (SYSCMD,"$command |");
      @command_results = <SYSCMD>;
   close (SYSCMD);
   $close_return_code = $?;
   if ( $close_return_code == 0 ) {
      $size=@command_results;
   }
   else { $size=0; }
}
#
#==============================================================================
#==============================================================================
#
#          Subroutine to determine if SKIPTEST is to be edited
#
#==============================================================================
#==============================================================================
sub test_to_replace_skiptest {
   if ( "$autogen_skiptest" eq "true" ) {
      if ( "$bypass_flag" eq "true" && 
      "$redundant_skiptest_commands" eq "false" ) {
         print "Automatically updating SKIPTEST flag within test stript...\n";
         &replace_skiptest;
      }
      else {
         if ( "$redundant_skiptest_commands" eq "true" ) {
            print "SKIPTEST flag multiply defined in $test_source test file.\n";
            print "Unable to edit redundant control flags.  Action abort.\n";
         }
         else {
            print "SKIPTEST flag not defined in $test_source test file.\n";
            print "Unable to edit non-existant control flag.  Action abort.\n";
         }
      }
   }
}
#
#==============================================================================
#==============================================================================
#
#          Subroutine to edit SKIPTEST variable
#
#==============================================================================
#==============================================================================
sub replace_skiptest {
#
#
# Clean up list of failed tests
   $failed_test_list = join(', ', sort( split(' ',$failed_test_list) ) );
#
#
# Generate archive file name
   ($dow, $mon, $mday, $time, $tz, $year) = split(' ', &ctime(time));
   $original_file = "$test_source.ORIG.$year$mon$mday";
   $updated_file = "$test_source";
#
#
# Display SKIPTEST header
   $old_bypass_list = $original_bypass_list;
   $new_bypass_list = $failed_test_list;
   $~ = SKIPTEST_MESSAGE;
   write;
#
#
# Determine which failed tests were not in the original series
   $additional_failed_tests = "";
   foreach $test_number ( split(', ',$failed_test_list) ) {
      if ( $original_bypass_list !~ m/$test_number,?/ ) { 
         if ( "$additional_failed_tests" eq "" ) {
            $additional_failed_tests = "$test_number";
         }
         else {
            $additional_failed_tests = "$additional_failed_tests $test_number";
         }
      }
   }
   $additional_passed_tests = "";
   foreach $test_number ( split(', ',$original_bypass_list) ) {
      if ( $failed_test_list !~ m/$test_number,?/ ) { 
         if ( "$additional_passed_tests" eq "" ) {
            $additional_passed_tests = "$test_number";
         }
         else {
            $additional_passed_tests = "$additional_passed_tests $test_number";
         }
      }
   }
#
#
# If no additional tests failed, display message
   if ( "$additional_failed_tests" eq "" && "$additional_passed_tests" eq "" ) {
      $test_step = "No changes.";
      $test_description = "All failed tests previously defined in SKIPTEST.";
      $test_state = "";
      $~ = SKIPTEST_DATA;
      write;
   }
#
#
# Copy original series to archive file
   $os_command = "/bin/cp $updated_file $original_file 1>/dev/null 2>&1";
   system "$os_command" || warn "   ERROR: backup copy failed: $!\n";
   $return_value = $? ;
#
#
# If archive create did not exit with error, perform update
   if ( "$return_value" eq "0" ) { 
#
#
# Open input and output files
      open (ORIGINAL_TEST,"<$original_file");
      open (EDITED_TEST,">$updated_file");
#
#
# Process each line from the input file
      while (<ORIGINAL_TEST>) {
         chomp;
#
#
# If SKIPTEST definition, replace
         if ( m/^SKIPTEST=/ ) {
            print EDITED_TEST "SKIPTEST=$failed_test_list;\n";
         }
#
#
# Else write out line to output file
         else {
            print EDITED_TEST "$_\n";
#
#
# If test step and test step matches additional bypassed test
            if ( "$additional_failed_tests" ne "None" ) {
               foreach $test_number ( split(' ',$additional_failed_tests) ) {
                  if ( $_ =~ /^$test_number/ ) { 
                     ($test_number, $description, @garbage) = split(/\;/,$_);
                     $test_step = "$test_number";
                     $test_description = "$description";
                     $test_state = "Fail";
                     $~ = SKIPTEST_DATA;
                     write;
                  }
               }
            }
            if ( "$additional_passed_tests" ne "None" ) {
               foreach $test_number ( split(' ',$additional_passed_tests) ) {
                  if ( $_ =~ /^$test_number/ ) { 
                     ($test_number, $description, @garbage) = split(/\;/,$_);
                     $test_step = "$test_number";
                     $test_description = "$description";
                     $test_state = "Pass";
                     $~ = SKIPTEST_DATA;
                     write;
                  }
               }
            }
         }
      }
#
#
# Close input and output files
      close (ORIGINAL_TEST);
      close (EDITED_TEST);
#
#
# Display trailer messages
      print "\n\nUpdate process completed.\n";
   }
#
#
# Else display abort message
   else {
      print "\n\nUpdate process aborted.\n";
   }
}
